(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: BoundCompliantSubstitutions_proofs.v *)
(* Description: Proofs about substitutions that match bounds *)

(** * Proofs about substitutions that match bounds *)

Require Import List.
Require Import Arith.
(* Require Import Relations. *)
Require Import Wf_nat.

Require Import MatchCompNat.
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Progress_proofs.v *)
Require Progress_proofs.
Module My_Progress_proofs := Progress_proofs.SetProgram(My_Program).
Import My_Progress_proofs.

Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)

Lemma Lifting_preserves_removeBounds:
  forall (t t': TypeS) (k n: nat),
    (removeBounds t t') ->
    (removeBounds (lift_at_depth_in_type t n k) t')
.

Proof.
intros. generalize k. clear k.
induction H; intros; simpl; constructor; auto.
Qed.

Lemma removeBounds_invariant_by_lifting:
  forall (t t': TypeS) (k n: nat),
    (removeBounds t t') ->
    (lift_at_depth_in_type t' n k) = t'
.

Proof.
intros. generalize k. clear k.
induction H.
intros. simpl. rewrite IHremoveBounds. trivial.
intros. simpl. rewrite IHremoveBounds. trivial.
intros. simpl. trivial.
intros. simpl. rewrite IHremoveBounds1. rewrite IHremoveBounds2. trivial.
Qed.

Lemma lift_of_some_inversion:
  forall (t N: TypeS) (n k: nat),
    (lift_at_depth_in_type t n k) = (Type_some N) ->
    exists N0: TypeS,
      t = (Type_some N0) /\
      N = (lift_at_depth_in_type N0 n k)
.

Proof.
intros. induction t; ( inversion H || inversion H; exists t; auto ).
Qed.

Lemma lift_of_cons_inversion:
  forall (t u us: TypeS) (n k: nat),
    (lift_at_depth_in_type t n k) = (Type_cons u us) ->
    exists u0: TypeS, exists us0: TypeS,
      t  = (Type_cons u0 us0) /\
      u  = (lift_at_depth_in_type u0 n k) /\
      us = (lift_at_depth_in_type us0 n k)
.

Proof.
intros. induction t; inversion H.
exists t1; exists t2; tauto.
Qed.

Lemma TypeS_at_of_lift_inversion:
  forall (t u: TypeS) (i n k: nat),
    (TypeS_at (lift_at_depth_in_type t n k) i u) ->
    exists u0: TypeS,
      (TypeS_at t i u0) /\
      u = (lift_at_depth_in_type u0 n k)
.

Proof.
cut (
  forall (t' u: TypeS) (i: nat),
    (TypeS_at t' i u) ->
    forall (t: TypeS) (n k: nat),
      t' = (lift_at_depth_in_type t n k) ->
      exists u0: TypeS,
        (TypeS_at t i u0) /\
        u = (lift_at_depth_in_type u0 n k)
).
intros. eapply H; eauto.
intros t' u i H. induction H.
(* *)
intros.
assert ((lift_at_depth_in_type t0 n k) = (Type_cons t ts)); auto. clear H.
destruct (lift_of_cons_inversion _ _ _ _ _ H0).
destruct H. destruct H. destruct H1. subst t. subst ts. subst t0.
exists x. split. constructor. trivial.
(* *)
intros.
assert ((lift_at_depth_in_type t0 n0 k) = (Type_cons t ts)); auto. clear H0.
destruct (lift_of_cons_inversion _ _ _ _ _ H1).
destruct H0. destruct H0. destruct H2. subst t. subst ts. subst t0.
destruct (IHTypeS_at x0 n0 k).  trivial. destruct H0.
exists x1. split. constructor; trivial. trivial.
Qed.

(********************************************************************)

Lemma Weakening_of_param_satisfaction:
  forall (D1 D2 D3: TypeEnv) (K P: TypeS),
    (ParamSat (D1 ++ D3) K P) ->
    (ParamSat (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
                (lift_at_depth_in_type K (length D2) (length D3))
                (lift_at_depth_in_type P (length D2) (length D3)))
.

Proof.
intros. inversion_clear H. simpl.
apply ParamSat_intro with (Pi' := Pi').
apply Lifting_preserves_removeBounds; trivial.
intros.
rewrite snoc_of_append. rewrite snoc_of_append.
replace Pi' with (lift_at_depth_in_type Pi' (length D2) (length D3)).
Focus 2. rewrite removeBounds_invariant_by_lifting with (t := Pi); trivial.
rewrite <- lift_of_snoc.
destruct (lift_of_some_inversion _ _ _ _ H). destruct H2. rename x into N0.
subst N. subst N_opt.
assert (Subtyping (snoc (D1 ++ D3) Pi') (Type_apply (lift_in_type K 1) (Type_var 0)) N0).
auto. clear H1. rewrite snoc_of_append in H2.
replace (S (length D3)) with (length (snoc D3 Pi')).
Focus 2. rewrite length_of_snoc. trivial.
replace (Type_apply
          (lift_in_type (lift_at_depth_in_type K (length D2) (length D3)) 1)
          (Type_var 0)) with
        (lift_at_depth_in_type (Type_apply (lift_in_type K 1) (Type_var 0))
          (length D2) (length (snoc D3 Pi'))).
apply Weakening_of_subtyping; trivial. simpl.
rewrite length_of_snoc.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt. Focus 2. omega.
pattern (length D3) at 2.
replace (length D3) with ((S (length D3)) - 1). Focus 2. omega.
rewrite Permuting_type_liftings_cor1. trivial. omega.
Qed.

Lemma Weakening_of_satisfaction:
  forall (D1 D2 D3: TypeEnv) (R Pi: TypeS),
    (SectionSat (D1 ++ D3) R Pi) ->
    (SectionSat (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
                (lift_at_depth_in_type R (length D2) (length D3))
                (lift_at_depth_in_type Pi (length D2) (length D3)))
.

Proof.
intros.
inversion_clear H. simpl.
eapply SectionSat_intro.
(* *)
apply Lifting_preserves_kind_of. eauto.
(* *)
apply Weakening_of_well_kindedness; eauto.
(* *)
unfold Forall_ParamSat. unfold Forall_ParamSat in H2.
intros.
rewrite <- Permuting_type_lifting_and_substitution_2_cor1 in H.
destruct (TypeS_at_of_lift_inversion _ _ _ _ _ H). clear H. destruct H3.
rename x into P0. subst P.
(* *)
change (Type_proj (lift_at_depth_in_type R (length D2) (length D3)) i) with
       (lift_at_depth_in_type (Type_proj R i) (length D2) (length D3)).
apply Weakening_of_param_satisfaction; auto.
Qed.

Lemma Weakening_of_satisfaction_by_the_right:
  forall (D D': TypeEnv) (R Pi: TypeS),
    (SectionSat D R Pi) ->
    (SectionSat (D ++ D')
                (lift_in_type R (length D'))
                (lift_in_type Pi (length D')))
.

Proof.
intros.
cut (
    (SectionSat (D ++ D' ++ (lift_in_typeEnv nil (length D')))
                (lift_at_depth_in_type R (length D') (length (nil (A := TypeS))))
                (lift_at_depth_in_type Pi (length D') (length (nil (A := TypeS)))))
).
unfold lift_in_typeEnv. simpl. rewrite <- app_nil_end. trivial.
apply Weakening_of_satisfaction. rewrite <- app_nil_end. trivial.
Qed.

(********************************************************************)

Fixpoint size_of_kind (k: KindS): nat :=
  match k with
    | (Kind_tuple k1)   => (size_of_kind k1) + 1
    | (Kind_constr k1)  => (size_of_kind k1) + 1
    | Kind_nil          => 0
    | (Kind_cons k1 k2) => (size_of_kind k1) + (size_of_kind k2) + 1
  end
.

Lemma typeSubst_of_class_type:
  forall (C: ClassSym) (R R': TypeS) (n: nat),
    (typeSubst_in_type (Class_type C R) n R') =
    (Class_type C (typeSubst_in_type R n R'))
.

Proof.
trivial.
Qed.

Lemma Substitution_in_closed_type_aux:
  forall (D: TypeEnv) (t R: TypeS) (c: TypeC) (k: nat),
    (Type_WK D t c) ->
    (typeSubst_in_type t ((length D) + k) R) = t
.

Proof.
intros. generalize k. clear k. induction H.
(* WK-proj *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-class *)
intros. simpl. trivial.
(* WK-fun *)
intro. simpl. rewrite IHType_WK1.
rewrite length_of_snoc in IHType_WK2. simpl in IHType_WK2. rewrite IHType_WK2. trivial.
(* WK-apply *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
(* WK-var *)
intro. simpl. destruct H. unfold typeSubst_in_typeVar.
rewrite match_comp_nat3_lt. trivial.
cut (X < (length D)). intro. omega. eapply TypeSection_at_lem9. apply H.
(* WK-tuple *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-param *)
intro. simpl. rewrite IHType_WK1.
rewrite length_of_snoc in IHType_WK2. simpl in IHType_WK2. rewrite IHType_WK2. trivial.
(* WK-section *)
intro. simpl.
rewrite length_of_snoc in IHType_WK. simpl in IHType_WK. rewrite IHType_WK. trivial.
(* WK-none *)
intro. simpl. trivial.
(* WK-some *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-param-nil *)
intro. simpl. trivial.
(* WK-param-cons *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
(* WK-constr-nil *)
intro. simpl. trivial.
(* WK-constr-cons *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
Qed.

(********************************************************************)

Lemma removeBounds_is_idempotent:
  forall (t t': TypeS),
    (removeBounds t t') ->
    (removeBounds t' t')
.

Proof.
intros. induction H.
apply removeBounds_param; trivial.
apply removeBounds_section; trivial.
apply removeBounds_param_nil.
apply removeBounds_param_cons; trivial.
Qed.

Lemma removeBounds_preserves_kind_of:
  forall (t t': TypeS) (k: KindS),
    (kind_of t k) ->
    (removeBounds t t') ->
    (kind_of t' k)
.

Proof.
intros. generalize t' H0. clear t' H0. induction H.
intros. inversion H0. apply Kind_of_section; eauto.
intros. inversion H0. apply Kind_of_param; eauto.
intros. inversion H0. apply Kind_of_paramList_nil.
intros. inversion H1. apply Kind_of_paramList_cons; eauto.
Qed.

Lemma Substitution_preserves_removeBounds:
  forall (t t' R: TypeS) (k: nat),
    (removeBounds t t') ->
    (removeBounds (typeSubst_in_type t k R) t')
.

Proof.
intros. generalize k. clear k. induction H.
intros. simpl. apply removeBounds_param; auto.
intros. simpl. apply removeBounds_section; auto.
intros. simpl. apply removeBounds_param_nil.
intros. simpl. apply removeBounds_param_cons; auto.
Qed.

Lemma removeBounds_then_substitution:
  forall (t t' R: TypeS) (k: nat),
    (removeBounds t t') ->
    (typeSubst_in_type t' k R) = t'
.

Proof.
intros. generalize k. clear k. induction H.
simpl. intros. rewrite IHremoveBounds. trivial.
simpl. intros. rewrite IHremoveBounds. trivial.
simpl. intros. trivial. 
simpl. intros. rewrite IHremoveBounds1. rewrite IHremoveBounds2. trivial.
Qed.

Lemma removeBounds_invariant_by_substitution:
  forall (t t' R: TypeS) (k: nat),
    (removeBounds t t') ->
    (typeSubst_in_type t' k R) = t'
.

Proof.
intros. generalize k. clear k.
induction H.
intros. simpl. rewrite IHremoveBounds. trivial.
intros. simpl. rewrite IHremoveBounds. trivial.
intros. simpl. trivial.
intros. simpl. rewrite IHremoveBounds1. rewrite IHremoveBounds2. trivial.
Qed.

(*************************************************************)

Lemma TypeParamList_ind:
  forall (D: TypeEnv) (Q: TypeS -> Prop),
    (Q Type_nil) ->
    (forall (P Ps: TypeS),
      (Type_WK D P TypeParam) ->
      (Type_WK D Ps TypeParamList) ->
      (Q Ps) -> 
      (Q (Type_cons P Ps))
    ) ->
    forall (Ps: TypeS),
      (Type_WK D Ps TypeParamList) ->
      (Q Ps)
.

Proof.
cut (
  forall (D: TypeEnv) (Q: TypeS -> Prop),
    (Q Type_nil) ->
    (forall (P Ps: TypeS),
      (Type_WK D P TypeParam) ->
      (Type_WK D Ps TypeParamList) ->
      (Q Ps) -> 
      (Q (Type_cons P Ps))
    ) ->
    forall (Ps: TypeS) (c: TypeC),
      (Type_WK D Ps c) ->
      c = TypeParamList ->
      (Q Ps)
).
intros. eapply H. trivial. apply H1. apply H2. trivial.
intros D Q IHnil IHcons Ps c H. induction H.
intro. discriminate H1.
intro. discriminate H0.
intro. discriminate H3.
intro. discriminate H1.
intro. discriminate H1.
intro. discriminate H0.
intro. discriminate H1.
intro. discriminate H0.
intro. discriminate H.
intro. discriminate H1.
intro. trivial.
intro. clear IHType_WK1.
apply IHcons. trivial. trivial. apply IHType_WK2. trivial. trivial.
intro. discriminate H.
intro. discriminate H1.
Qed.

(**********************************************************************)

Lemma kind_of_params_inversion:
  forall (Ps P: TypeS) (k ks: KindS) (i: nat),
    (kind_of Ps ks) ->
    (TypeS_at Ps i P) ->
    (KindS_at ks i k) ->
    (kind_of P k)
.

Proof.
intros. generalize k ks H H1. clear k ks H H1. induction H0.
(* *)
intros. inversion H. subst ks. inversion H1. rewrite <- H4. trivial.
(* *)
intros. inversion H. subst ks. inversion H1. eapply IHTypeS_at; eauto.
Qed.

Lemma kind_of_params_inversion_variant1:
  forall (Ps P: TypeS) (ks: KindS) (i: nat),
    (kind_of Ps ks) ->
    (TypeS_at Ps i P) ->
    exists k: KindS,
      (KindS_at ks i k) /\
      (kind_of P k)
.

Proof.
intros. generalize ks H. clear ks H. induction H0.
(* *)
intros. inversion H. subst ks. exists k. split. constructor. trivial.
(* *)
intros. inversion H. subst ks.
destruct (IHTypeS_at _ H5). destruct H4. rename x into k0.
exists k0. split. constructor; trivial. trivial.
Qed.

(***********************************************************************)

Lemma removeBounds_of_params_inversion:
  forall (Ps Ps' P': TypeS) (i: nat),
    (removeBounds Ps Ps') ->
    (TypeS_at Ps' i P') ->
    exists P: TypeS,
      (TypeS_at Ps i P) /\
      (removeBounds P P')
.

Proof.
intros. generalize Ps H. clear Ps H. induction H0.
(* *)
intros.
inversion H. subst Ps.
exists P. split. constructor. trivial.
(* *)
intros.
inversion H. subst Ps.
destruct (IHTypeS_at _ H5). destruct H3. rename x into P0.
exists P0. split. constructor. trivial. trivial.
Qed.

Lemma removeBounds_and_kind_of_inversion:
  forall (t t': TypeS) (k: KindS),
    (removeBounds t t') ->
    (kind_of t' k) ->
    (kind_of t k)
.

Proof.
intros. generalize k H0. clear k H0. induction H.
(* *)
intros. inversion H0. constructor. auto.
(* *)
intros. inversion H0. constructor. auto.
(* *)
intros. trivial.
(* *)
intros. inversion H1. constructor; auto.
Qed.

(***********************************************************************)

Lemma kind_equality_implies_param_satisfaction:
  forall (D: TypeEnv) (K P P' Pi N_opt: TypeS) (k: KindS),
    P = (Type_param Pi N_opt) ->
    (Type_WK D K (TypeConstr k)) ->
    (removeBounds P P') ->
    (kind_of P k) ->
    (ParamSat D K P')
.

Proof.
intros.
rewrite H in H1. inversion H1. clear Pi0 T H3 H5.
apply ParamSat_intro with (Pi' := Pi'). 
eapply removeBounds_is_idempotent; eauto.
intros. discriminate H3.
Qed.

Theorem kind_equality_implies_poly_satisfaction:
  forall (D: TypeEnv) (R Pi Pi': TypeS) (k: KindS),
    (Type_WK D R (TypeTuple k)) ->
    (Type_WK D Pi TypeSection) ->
    (kind_of Pi k) ->
    (removeBounds Pi Pi') ->
    (SectionSat D R Pi')
.

Proof.
intros.
(* *)
inversion H0. subst Pi. clear D0 H4. clear H0.
(* *)
inversion H1. rename k0 into ks. subst k. clear Ps0 H0. clear H1.
(* *)
inversion H2. subst Pi'.  clear Ps0 H0. clear H2.
(* *)
eapply SectionSat_intro. 
(* *)
eapply removeBounds_preserves_kind_of; eauto.
(* *)
trivial.
(* *)
unfold Forall_ParamSat. intros. rename P into P'.
(* *)
assert (Hy0: exists k : KindS, (KindS_at ks i k) /\ (kind_of P' k)).
assert (Hy1: kind_of (typeSubst_in_type Ps' 0 R) ks).
apply Substitution_preserves_kind_of.
eapply removeBounds_preserves_kind_of. eauto. trivial.
eapply kind_of_params_inversion_variant1. apply Hy1. trivial.
destruct Hy0. destruct H2. rename x into k.
(* *)
assert (Hy2: Type_WK D (Type_proj R i) (TypeConstr k)).
econstructor; eauto.
(* *)
assert (Hy3: TypeS_at Ps' i P').
rewrite <- (removeBounds_invariant_by_substitution _ _ R 0 H1). trivial.
(* *)
assert (Hy4: exists P: TypeS, (TypeS_at Ps i P) /\ (removeBounds P P')).
eapply removeBounds_of_params_inversion; eauto.
destruct Hy4. destruct H6. rename x into P.
(* *)
assert (Hy5: Type_WK (snoc D (Type_section Ps)) P TypeParam).
eapply TypeParamList_WK_inversion; eauto.
inversion Hy5. clear D0 H10.
(* *)
eapply kind_equality_implies_param_satisfaction.
(* *)
symmetry. eauto.
(* *)
eauto.
(* *)
trivial.
(* *)
eapply removeBounds_and_kind_of_inversion; eauto.
Qed.

(**********************************************************************)

Lemma TypeEnv_WK_inversion:
  forall (D: TypeEnv) (i: nat) (Pi: TypeS),
    (TypeEnv_WK D) ->
    (TypeSection_at D i Pi) ->
    (Type_WK D (lift_in_type Pi (S i)) TypeSection)
.

Proof.
intros. generalize H. clear H. induction H0.
(* *)
intros. inversion H. absurd ((snoc D Pi) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H0). clear H0. rewrite <- H4. 
apply Weakening_of_well_kindedness_by_the_right_snoc. trivial.
(* *)
intros. inversion H. absurd ((snoc D Pi) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H1). clear H1. 
replace (lift_in_type Pi' (S (S n))) with (lift_in_type (lift_in_type Pi' (S n)) 1).
apply Weakening_of_well_kindedness_by_the_right_snoc.
rewrite H4. apply IHTypeSection_at. rewrite <- H4. trivial.
replace (S (S n)) with ((S n) + 1).
apply Successive_type_liftings. simpl. omega.
Qed.

Lemma TypeEnv_WK_inversion_cor1:
  forall (D: TypeEnv) (X: TypeVar) (Pi: TypeS),
    (TypeEnv_WK D) ->
    (decl_of_typeVar D X Pi) ->
    (Type_WK D Pi TypeSection)
.

Proof.
intros. inversion H0. apply TypeEnv_WK_inversion; trivial.
Qed.

(**********************************************************************)

Lemma Var_expansion_preserves_well_kindedness_aux:
  forall (Ps Pi N_opt: TypeS) (i: nat) (k ks: KindS),
    (TypeS_at Ps i (Type_param Pi N_opt)) ->
    (kind_of Ps ks) ->
    (KindS_at ks i (Kind_constr k)) ->
    (kind_of Pi k)
.

Proof.
intros. generalize Pi N_opt i k H H1. clear Pi N_opt i k H H1. induction H0.
(* *)
intros. inversion H1.
(* *)
intros. inversion H1.
(* *)
intros. inversion H1.
(* *)
clear IHkind_of1.
intros. inversion H.
 (* *)
 rewrite <- H0 in H1. clear t ts H2 H3 H0. inversion H1. clear k1 ks0 H2 H3.
 rewrite H4 in H0_. inversion H0_. clear Pi0 T_opt H2 H5. rewrite H0 in H3.
 injection H3. intro. rewrite <- H2. trivial.
 (* *)
 rewrite <- H2 in H1. clear t ts u H0 H3 H2 H4. inversion H1.
 eapply IHkind_of2. apply H5. trivial.
Qed.

Theorem Var_expansion_preserves_well_kindedness:
  forall (D: TypeEnv) (X: TypeVar) (R Ps Pi N: TypeS) (i: nat),
    (TypeEnv_WK D) ->
    (decl_of_typeVar D X (Type_section Ps)) ->
    (TypeS_at (typeSubst_in_type Ps 0 (Type_var X)) i (Type_param Pi (Type_some N))) ->
    (Type_WK D (Var_type X i R) Type_) ->
    (Type_WK D (typeSubst_in_type N 0 R) Type_)
.

Proof.
intros.
(* *)
inversion H2. clear D0 K R0 H5 H3 H4. inversion H6. clear D0 R0 i0 k' H5 H3 H4 H8.
inversion H9. clear D0 X0 k1 H5 H3 H4.
(* *)
assert (Pi0 = (Type_section Ps)).
eapply decl_of_typeVar_is_a_partial_function; eauto.
rewrite H3 in H11. clear Pi0 H8 H3.
(* *)
assert (Type_WK D (Type_section Ps) TypeSection).
apply TypeEnv_WK_inversion_cor1 with (X := X); trivial.
(* *)
inversion H3. clear H3 D0 Ps0 H5 H4.
(* *)
assert (Type_WK D (typeSubst_in_type Ps 0 (Type_var X)) TypeParamList).
eapply Substitution_preserves_well_kindedness_snoc. apply H9. apply H11. trivial.
(* *)
assert (Type_WK D (Type_param Pi (Type_some N)) TypeParam).
eapply TypeParamList_WK_inversion; eauto.
(* *)
inversion H4. clear H4 D0 Pi0 N_opt H13 H5 H12. inversion H15. clear H15 D0 N0 H12 H4.
(* *)
eapply Substitution_preserves_well_kindedness_snoc. apply H7.
eapply Var_expansion_preserves_well_kindedness_aux. apply H1.
apply Substitution_preserves_kind_of. inversion H11. apply H15. trivial.
trivial.
Qed.

(**********************************************************************)

Lemma TypeEnv_WK_inversion_cor2:
  forall (D1 D2: TypeEnv) (Pi: TypeS),
    (TypeEnv_WK (D1 ++ (Pi :: D2))) ->
    (Type_WK D1 Pi TypeSection)
.

Proof.
intros. generalize H. clear H. pattern D2. apply (snoc_ind TypeS).
(* *)
intros. change (D1 ++ Pi :: nil) with (snoc D1 Pi) in H.
inversion H. absurd ((snoc D1 Pi) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H0). rewrite <- H3. rewrite <- H4. trivial.
(* *)
intros.
replace (D1 ++ Pi :: snoc xs x) with (snoc (D1 ++ Pi :: xs) x) in H0. Focus 2.
rewrite snoc_of_append. rewrite snoc_of_cons. trivial.
inversion H0. absurd ((snoc (D1 ++ Pi :: xs) x) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H1). apply H. rewrite <- H4. trivial.
Qed.

(**********************************************************************)

Lemma KindS_at_and_size:
  forall (k ks: KindS) (i: nat),
    (KindS_at ks i k) ->
    (size_of_kind k) < (size_of_kind ks)
.

Proof.
intros. induction H.
simpl. omega.
simpl. omega.
Qed.

(**********************************************************************)

Lemma Substitution_in_well_kinded_type_env:
  forall (D1 D2: TypeEnv) (Pi R: TypeS) (k: KindS),
    (TypeEnv_WK (D1 ++ Pi :: D2)) ->
    (Type_WK D1 R (TypeTuple k)) ->
    (kind_of Pi k) ->
    (TypeEnv_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R)))
.

Proof.
intros. generalize H. clear H. pattern D2. apply (snoc_ind TypeS).
(* nil *)
intros. simpl. rewrite <- app_nil_end. fold (snoc D1 Pi) in H. inversion H.
absurd ((snoc D1 Pi) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H2). rewrite <- H5. trivial.
(* snoc *)
intros. rewrite <- snoc_of_cons in H2. rewrite <- snoc_of_append in H2.
inversion H2. absurd ((snoc (D1 ++ Pi :: xs) x) = nil). apply snoc_ind_aux_3. auto.
destruct (snoc_inject _ _ _ _ _ H3). subst D. subst x. clear H2 H3.
rewrite typeSubst_of_snoc. rewrite <- snoc_of_append.
constructor. tauto. simpl.
eapply Type_substitution_preserves_well_kindedness; eauto.
Qed.

Lemma removeBounds_preserves_well_kindedness:
  forall (D: TypeEnv) (t t': TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (removeBounds t t') ->
    forall (D': TypeEnv),
      (Type_WK D' t' c)
.

Proof.
intros. generalize t' H0 D'. clear t' H0 D'. induction H.
intros. inversion H1.
intros. inversion H0.
intros. inversion H3.
intros. inversion H1.
intros. inversion H1.
intros. inversion H0.
intros. inversion H1. constructor. auto. constructor.
intros. inversion H0. constructor. auto.
intros. inversion H0.
intros. inversion H1.
intros. inversion H0. constructor.
intros. inversion H1. constructor; auto.
intros. inversion H0. constructor.
intros. inversion H1. constructor; auto.
Qed.

(**********************************************************************)

Lemma Type_substitution_preserves_subtyping_aux1:
  forall (D1 D2: TypeEnv) (R Pi': TypeS) (i: nat) (k0 k1: KindS),
    (Type_WK D1 R (TypeTuple (Kind_tuple k1))) ->
    (KindS_at k1 i (Kind_constr k0)) ->
    (kind_of Pi' k0) ->
    (Type_WK (snoc (D1 ++ typeSubst_in_typeEnv D2 0 R) Pi')
             (Type_apply (lift_in_type (Type_proj (lift_in_type R (length D2)) i) 1)
             (Type_var 0)) Type_)
.

Proof.
intros. econstructor.
(* *)
apply Weakening_of_well_kindedness_by_the_right_snoc.
econstructor.
replace (length D2) with (length (typeSubst_in_typeEnv D2 0 R)).
Focus 2. apply length_of_typeSubst.
apply Weakening_of_well_kindedness_by_the_right. eauto. eauto.
(* *)
unfold snoc.
apply Weakening_of_well_kindedness_by_the_left.
econstructor. constructor. rewrite <- snoc_nil. constructor.
unfold lift_in_type. apply Lifting_preserves_kind_of. trivial.
Qed.

(*********************************************************************)

Theorem Type_substitution_preserves_subtyping:
  forall (D1 D2: TypeEnv) (Pi R T U: TypeS) (k: KindS),
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK (D1 ++ (Pi :: D2))) ->
    (Subtyping (D1 ++ (Pi :: D2)) T U) ->
    (Type_WK (D1 ++ (Pi :: D2)) T Type_) ->
    (Type_WK (D1 ++ (Pi :: D2)) U Type_) ->
    (Subtyping (D1 ++ (typeSubst_in_typeEnv D2 0 R))
               (typeSubst_in_type T (length D2) R)
               (typeSubst_in_type U (length D2) R))
.

Proof.
cut (
  forall (n: nat) (k: KindS),
    n = (size_of_kind k) ->
    forall (D1: TypeEnv) (Pi R: TypeS),
      (Type_WK D1 R (TypeTuple k)) ->
      (SectionSat D1 R Pi) ->
      forall (D: TypeEnv) (T U: TypeS),
        (Subtyping D T U) ->
        forall (D2: TypeEnv),
          D = (D1 ++ (Pi :: D2)) ->
          (Type_WK D T Type_) ->
          (Type_WK D U Type_) ->
          (TypeEnv_WK D) ->
          (Subtyping (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                     (typeSubst_in_type T (length D2) R)
                     (typeSubst_in_type U (length D2) R))
).
intros. eapply H; eauto.

intro n. induction (lt_wf n).
intros k H1 D1 Pi R H2 H3 D T U H4. induction H4.

(* S-Refl *)
clear H0. intros. apply S_Refl.

(* S-Trans *)
clear H0. intros. rename H7 into Hy.  eapply S_Trans.
eapply Type_substitution_preserves_well_kindedness
with (Pi := Pi).
apply H2. inversion H3. replace k with (Kind_tuple ks). 
apply Kind_of_section. trivial.
cut ((TypeTuple (Kind_tuple ks)) = (TypeTuple k)). intro. injection H11. tauto.
eapply Uniqueness_of_well_kindedness. apply H8. trivial.
apply SameCategory_TypeTuple.
apply H4. trivial.
apply IHSubtyping1; trivial.
apply IHSubtyping2; trivial.

(* S-Var *)
intros. rename H10 into Hy. simpl. destruct (lt_eq_lt_dec X (length D2)). destruct s.
 (* X < |D2| *)
clear H0.
rewrite H7 in H4.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
fold (Var_type X i (typeSubst_in_type R0 (length D2) R)).
eapply S_Var with (Ps := (typeSubst_in_type Ps (S (length D2)) R))
                  (Pi := (typeSubst_in_type Pi0 (length D2) R))
                  (N  := (typeSubst_in_type N (S (length D2)) R)).
change (Type_section (typeSubst_in_type Ps (S (length D2)) R)) with
       (typeSubst_in_type (Type_section Ps) (length D2) R).
eapply Declaration_of_type_variable_2; eauto. trivial.
replace (Type_var X) with (typeSubst_in_type (Type_var X) (length D2) R).
change (S (length D2)) with (S (0 + (length D2))).
rewrite <- Permuting_type_substitutions. simpl.
change (Type_param (typeSubst_in_type Pi0 (length D2) R)
        (Type_some (typeSubst_in_type N (S (length D2)) R))) with
        (typeSubst_in_type (Type_param Pi0 (Type_some N)) (length D2) R).
apply TypeS_at_of_typeSubst; trivial.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt. trivial. trivial.
change (S (length D2)) with (S (0 + (length D2))).
rewrite <- Permuting_type_substitutions. simpl.
apply IHSubtyping. trivial. clear IHSubtyping.
rewrite H7.
eapply Var_expansion_preserves_well_kindedness.  rewrite <- H7. trivial.
apply H4. apply H5. rewrite <- H7. trivial.
trivial. trivial. trivial.

 (* X = |D2| *)
rename H0 into IH. (* clear IH. *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq. Focus 2. trivial.
rename N' into U. subst D. subst X.
 (* *)
 assert (Hy0: Type_WK (D1 ++ Pi :: D2) (typeSubst_in_type N 0 R0) Type_).
 eapply Var_expansion_preserves_well_kindedness with (X := (length D2)); eauto.
 (* *)
 assert (Hy1: Subtyping (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                        (typeSubst_in_type (typeSubst_in_type
                         N (S (length D2)) R) 0 (typeSubst_in_type R0 (length D2) R))
                        (typeSubst_in_type U (length D2) R)).
 rewrite <- Permuting_type_substitutions_cor1.
 apply IHSubtyping; auto. clear IHSubtyping.
 (* *)
 assert (Hy2: Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                      (typeSubst_in_type (typeSubst_in_type
                       N (S (length D2)) R) 0 (typeSubst_in_type R0 (length D2) R))
                       Type_).
 rewrite <- Permuting_type_substitutions_cor1.
 eapply Type_substitution_preserves_well_kindedness. eauto.
 eapply Satisfaction_implies_same_kind. eauto. eauto. eauto. trivial.
 (* *)
 assert (Hy3: (Type_section Ps) = (lift_in_type Pi (S (length D2)))).
 eapply decl_of_typeVar_is_a_partial_function. apply H4.
 apply Declaration_of_type_variable_3.
 (* *) 
 assert (Hy4: Type_WK D1 Pi TypeSection).
 eapply TypeEnv_WK_inversion_cor2; eauto.
 inversion Hy4. clear D H0 H7. subst Pi.
 (* *)
 unfold lift_in_type in Hy3. simpl in Hy3. injection Hy3. clear Hy3. intros Hy5.
 (* *)
 assert (Hy6:
   (typeSubst_in_type (typeSubst_in_type Ps 0 (Type_var (length D2))) (length D2) R) =
   (typeSubst_in_type (lift_at_depth_in_type Ps0 (length D2) 1)
                      0 (lift_in_type R (length D2)))
 ).
 rewrite Permuting_type_substitutions_cor1.
 simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq. Focus 2. trivial.
 rewrite Hy5.
 replace (typeSubst_in_type (lift_at_depth_in_type Ps0 (S (length D2)) 1)
                           (S (length D2)) R) with
        (typeSubst_in_type (lift_at_depth_in_type Ps0 (S ((length D2) + 0)) 1)
                           ((length D2) + 1) R).
 Focus 2. replace (length D2 + 0) with (length D2).
 replace (length D2 + 1) with (S (length D2)). trivial. omega. omega.
 rewrite Type_lifting_stronger_than_substitution_aux. rewrite plus_comm. trivial.
 (* *)
 assert (Hy7: SectionSat (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                         (lift_in_type R (length D2))
                         (Type_section (lift_at_depth_in_type Ps0 (length D2) 1))).
 change (Type_section (lift_at_depth_in_type Ps0 (length D2) 1)) with
        (lift_in_type (Type_section Ps0) (length D2)).
 replace (length D2) with (length (typeSubst_in_typeEnv D2 0 R)).
 Focus 2. apply length_of_typeSubst.
 apply Weakening_of_satisfaction_by_the_right. trivial.
 (* *)
 assert (Hy8: (TypeS_at (typeSubst_in_type (lift_at_depth_in_type Ps0 (length D2) 1)
                                           0 (lift_in_type R (length D2))) i
              (Type_param (typeSubst_in_type Pi0 (length D2) R)
                          (Type_some (typeSubst_in_type N (S (length D2)) R))))).
 rewrite <- Hy6.
 change (Type_param (typeSubst_in_type Pi0 (length D2) R)
                          (Type_some (typeSubst_in_type N (S (length D2)) R))) with
        (typeSubst_in_type (Type_param Pi0 (Type_some N)) (length D2) R).
 apply TypeS_at_of_typeSubst. trivial.
 (* *)
 assert (Hy9: ParamSat (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                       (Type_proj (lift_in_type R (length D2)) i)
                       (Type_param (typeSubst_in_type Pi0 (length D2) R)
                          (Type_some (typeSubst_in_type N (S (length D2)) R)))).
 inversion Hy7. unfold Forall_ParamSat in H11. auto.
 clear Hy6. rename Hy7 into Hy21_0. rename Hy8 into Hy21_1.
 (* *)
 inversion Hy9. clear Pi N_opt H0 H7.
 assert (Hy10: Subtyping (snoc (D1 ++ typeSubst_in_typeEnv D2 0 R) Pi')
               (Type_apply
               (lift_in_type (Type_proj (lift_in_type R (length D2)) i) 1)
               (Type_var 0)) (typeSubst_in_type N (S (length D2)) R)). auto.
 clear Hy9 H11.
 (* *)
 inversion H8. clear D K R1 H11 H0 H7. rename H13 into Hy11.
 inversion H12. clear H12 D R1 i0 k' H11 H0 H7 H13.
 inversion H14. clear D X k2 H11 H0 H7 H14.
 assert (Hy12: Pi =
   (Type_section (lift_at_depth_in_type Ps0 (S (length D2)) 1))).
 rewrite <- Hy5.
 eapply decl_of_typeVar_is_a_partial_function; eauto.
 clear H12. subst Pi.
 inversion H13. clear H13 Ps1 k2 H0 H7.
(* *)
 assert (Hy13: Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                       (typeSubst_in_type R0 (length D2) R) (TypeTuple k0)).
 eapply Type_substitution_preserves_well_kindedness. eauto.
 eapply Satisfaction_implies_same_kind; eauto. eauto. eauto.
 clear Hy11.
 (* *)
 assert (Hy14: exists kPi: KindS, (kind_of (Type_section Ps0) kPi)).
 eapply wellKinded_implies_kind_of_exists; eauto.
 destruct Hy14. rename x0 into kPi. rename H0 into Hy14.
 (* *)
 assert (Hy15: (kind_of (Type_section Ps0) (Kind_tuple k1))).
 replace (Kind_tuple k1) with kPi. trivial. symmetry.
 eapply kind_of_is_a_partial_function.
 apply Kind_of_section. eauto.
 change (Type_section (lift_at_depth_in_type Ps0 (S (length D2)) 1)) with
        (lift_at_depth_in_type (Type_section Ps0) (S (length D2)) 0).
 apply Lifting_preserves_kind_of; trivial. clear kPi Hy14.
 (* *)
 assert (Hy18: kind_of Pi0 k0).
 cut (kind_of (Type_param Pi0 (Type_some N)) (Kind_constr k0)).
 intros. inversion H0. trivial.
 eapply kind_of_params_inversion with
   (Ps := ((typeSubst_in_type Ps 0 (Type_var (length D2)))))
   (ks := k1).
 apply Substitution_preserves_kind_of. rewrite Hy5.
 apply Lifting_preserves_kind_of. inversion Hy15; trivial.
 apply H5. trivial.
 (* *)
 assert (Hy19: kind_of Pi' k0).
 apply removeBounds_preserves_kind_of with (t := (typeSubst_in_type Pi0 (length D2) R)).
 apply Substitution_preserves_kind_of; trivial. trivial.
 (* *)
 assert (Hy20_0: k = (Kind_tuple k1)).
 symmetry.
 eapply kind_of_is_a_partial_function. apply Hy15.
 eapply Satisfaction_implies_same_kind. apply H2. trivial.
 (* *)
 assert (Hy20: (size_of_kind k0) < (size_of_kind k)).
 rewrite Hy20_0.
 simpl.
 cut ((size_of_kind k0) + 1 < (size_of_kind k1)). intros. omega.
 change (size_of_kind k0 + 1) with (size_of_kind (Kind_constr k0)).
 eapply KindS_at_and_size; eauto.
 (* *)
 assert (Hy21_2: Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                 (lift_in_type R (length D2)) (TypeTuple k)).
 rewrite <- length_of_typeSubst with (X := 0) (R := R).
 apply Weakening_of_well_kindedness_by_the_right. eauto.
 (* *)
 assert (Hy21_3: Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                       (Type_param (typeSubst_in_type Pi0 (length D2) R)
                       (Type_some (typeSubst_in_type N (S (length D2)) R)))
                       TypeParam).
 eapply TypeParamList_WK_inversion with
   (Ps := (typeSubst_in_type (lift_at_depth_in_type Ps0 (length D2) 1) 0
                             (lift_in_type R (length D2)))).
 eapply Substitution_preserves_well_kindedness_snoc with 
   (Pi := (lift_in_type (Type_section Ps0) (length D2))).
 eauto. unfold lift_in_type. apply Lifting_preserves_kind_of. rewrite Hy20_0. trivial.
 rewrite <- length_of_typeSubst with (X := 0) (R := R).
 unfold snoc. rewrite app_ass.
 change ((lift_in_type (Type_section Ps0) (length (typeSubst_in_typeEnv D2 0 R))) :: nil)
   with
   (lift_in_typeEnv ((Type_section Ps0) :: nil) (length (typeSubst_in_typeEnv D2 0 R))).
 change 1 with (length ((Type_section Ps0) :: nil)).
 apply Weakening_of_well_kindedness. inversion Hy4. trivial.
 eauto.

 (* *)
 assert (Hy21: Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                       (typeSubst_in_type Pi0 (length D2) R) TypeSection).
 inversion Hy21_3. trivial.
 (* *)
 assert (Hy22: SectionSat (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                          (typeSubst_in_type R0 (length D2) R) Pi').
 eapply kind_equality_implies_poly_satisfaction with
   (Pi := (typeSubst_in_type Pi0 (length D2) R)).
 eauto. trivial. apply Substitution_preserves_kind_of. trivial. trivial.
 (* *)
 assert (Hy23_0: Type_WK (snoc (D1 ++ typeSubst_in_typeEnv D2 0 R) Pi')
                         (typeSubst_in_type N (S (length D2)) R) Type_).
 unfold snoc.
 eapply Well_kindedness_does_not_depend_on_bounds with
   (Pi := (typeSubst_in_type Pi0 (length D2) R)).
 apply Substitution_preserves_kind_of. eauto. trivial.
 inversion Hy21_3. inversion H14. trivial.
 (* *)
 assert (Hy23: Subtyping (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                (Type_apply (Type_proj (lift_in_type R (length D2)) i)
                            (typeSubst_in_type R0 (length D2) R))
                (typeSubst_in_type (typeSubst_in_type
                 N (S (length D2)) R) 0 (typeSubst_in_type R0 (length D2) R))).
 replace (Type_apply (Type_proj (lift_in_type R (length D2)) i)
                     (typeSubst_in_type R0 (length D2) R)) with
         (typeSubst_in_type
            (Type_apply (lift_in_type (Type_proj (lift_in_type R (length D2)) i) 1)
            (Type_var 0)) 0 (typeSubst_in_type R0 (length D2) R)).
  (* *)
  replace (D1 ++ typeSubst_in_typeEnv D2 0 R) with
         ((D1 ++ typeSubst_in_typeEnv D2 0 R) ++
          (typeSubst_in_typeEnv nil 0 (typeSubst_in_type R0 (length D2) R))).
 change (Subtyping
     ((D1 ++ typeSubst_in_typeEnv D2 0 R) ++
      typeSubst_in_typeEnv nil 0 (typeSubst_in_type R0 (length D2) R))
     (typeSubst_in_type
        (Type_apply
           (lift_in_type (Type_proj (lift_in_type R (length D2)) i) 1)
           (Type_var 0)) (length (nil (A := TypeS))) (typeSubst_in_type R0 (length D2) R))
     (typeSubst_in_type (typeSubst_in_type N (S (length D2)) R)
       (length (nil (A := TypeS))) (typeSubst_in_type R0 (length D2) R))).
  subst x.
  eapply (IH _ Hy20 k0). trivial. trivial. eauto. apply Hy10. trivial.
  eapply Type_substitution_preserves_subtyping_aux1; eauto.
  rewrite <- Hy20_0. trivial.
  trivial.
  constructor. eapply Substitution_in_well_kinded_type_env; eauto.
  rewrite Hy20_0. trivial.
  eapply removeBounds_preserves_well_kindedness with
    (t := (typeSubst_in_type Pi0 (length D2) R)); eauto.
  simpl. rewrite <- app_nil_end. trivial.
  (* *)
 simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
 unfold lift_in_type.
 rewrite lift_in_type_by_zero.
 change (typeSubst_in_type
           (lift_at_depth_in_type (lift_at_depth_in_type R (length D2) 0) 1 0)
           0 (typeSubst_in_type R0 (length D2) R)) with
         (typeSubst_in_type
           (lift_in_type (lift_at_depth_in_type R (length D2) 0) (S (0 + 0)))
           0 (typeSubst_in_type R0 (length D2) R)).
  rewrite Type_lifting_stronger_than_substitution. simpl.
 unfold lift_in_type. rewrite lift_in_type_by_zero. trivial.
 trivial.
 (* *)
 eapply S_Trans. apply Hy2. trivial. trivial.

 (* X > |D2| *)
clear H0.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
fold (Var_type (pred X) i (typeSubst_in_type R0 (length D2) R)).
rewrite H7 in H4.
eapply S_Var with (Ps := (typeSubst_in_type Ps (S (length D2)) R))
                  (N := (typeSubst_in_type N (S (length D2)) R))
                  (Pi := (typeSubst_in_type Pi0 (length D2) R)).
change (Type_section (typeSubst_in_type Ps (S (length D2)) R)) with
        (typeSubst_in_type (Type_section Ps) (length D2) R).
eapply Declaration_of_type_variable_1_bis. trivial. apply H4.
(* begin *)
replace (Type_var (pred X)) with (typeSubst_in_type (Type_var X) (length D2) R).
change (S (length D2)) with (S (0 + (length D2))).
rewrite <- Permuting_type_substitutions. simpl.
change  (Type_param (typeSubst_in_type Pi0 (length D2) R)
        (Type_some (typeSubst_in_type N (S (length D2)) R))) with
        (typeSubst_in_type (Type_param Pi0 (Type_some N)) (length D2) R).
apply TypeS_at_of_typeSubst. trivial.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt. trivial. trivial.
(* end *)
change (S (length D2)) with (S (0 + length D2)).
rewrite <- Permuting_type_substitutions.
simpl. apply IHSubtyping. trivial.
(* *)
rewrite H7.
eapply Var_expansion_preserves_well_kindedness.  rewrite <- H7. trivial.
apply H4. apply H5. rewrite <- H7. trivial.
(* *)
trivial. trivial. trivial.

(* S-Class *)
clear H0. intros. rename H8 into Hy.
rewrite typeSubst_of_class_type. eapply S_Class. apply H4.
replace (typeSubst_in_type N 0 (typeSubst_in_type R0 (length D2) R)) with
  (typeSubst_in_type (typeSubst_in_type N 0 R0) (length D2) R).
apply IHSubtyping. trivial.
eapply Class_type_expansion_preserves_well_kindedness. apply H6. trivial. trivial.
trivial.
pattern (length D2) at 1. replace (length D2) with (0 + (length D2)).
rewrite Permuting_type_substitutions. simpl.
replace (S (length D2)) with ((length ((classTypeParams C) :: nil)) + (length D2)).
rewrite Substitution_in_closed_type_aux with (c := Type_). trivial.
destruct (classes_are_well_kinded C). unfold Pi0 in H9.
destruct (H9 _ H4). tauto.
simpl. trivial. trivial.

(* S-Beta-Left *)
clear H0. intros. eapply S_Beta_Left. apply Substitution_in_reduction. apply H4.
apply IHSubtyping. trivial.
eapply Well_kindedness_is_preserved_by_type_reduction. apply H6. trivial. trivial. trivial.

(* S-Beta-Right *)
clear H0. intros. eapply S_Beta_Right. apply Substitution_in_reduction. apply H4.
apply IHSubtyping. trivial. trivial.
eapply Well_kindedness_is_preserved_by_type_reduction. apply H7. trivial. trivial.
Qed.

Theorem Type_substitution_preserves_subtyping_snoc:
  forall (D1: TypeEnv) (Pi R T U: TypeS) (k: KindS),
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK (snoc D1 Pi)) ->
    (Subtyping (snoc D1 Pi) T U) ->
    (Type_WK (snoc D1 Pi) T Type_) ->
    (Type_WK (snoc D1 Pi) U Type_) ->
    (Subtyping D1 (typeSubst_in_type T 0 R) (typeSubst_in_type U 0 R))
.

Proof.
intros.
cut (
    (Subtyping (D1 ++ (typeSubst_in_typeEnv nil 0 R))
               (typeSubst_in_type T (length (nil (A := TypeS))) R)
               (typeSubst_in_type U (length (nil (A := TypeS))) R))
).
simpl. rewrite <- app_nil_end. trivial.
eapply Type_substitution_preserves_subtyping; eauto.
Qed.

Lemma Type_substitution_preserves_subtyping_cor1:
  forall (Pi1 Pi2 R T U: TypeS) (k: KindS),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi1) ->
    let D := (Pi1 :: Pi2 :: nil) in
    (TypeEnv_WK D) ->
    (Type_WK D T Type_) ->
    (Type_WK D U Type_) ->
    (Subtyping D T U) ->
    (Subtyping ((typeSubst_in_type Pi2 0 R) :: nil)
               (typeSubst_in_type T 1 R)
               (typeSubst_in_type U 1 R))
.

Proof.
intros.
cut (
    (Subtyping (nil ++ (typeSubst_in_typeEnv (Pi2 :: nil) 0 R))
               (typeSubst_in_type T (length (Pi2 :: nil)) R)
               (typeSubst_in_type U (length (Pi2 :: nil)) R))
).
simpl. tauto.
eapply Type_substitution_preserves_subtyping; eauto.
Qed.

Lemma Type_substitution_preserves_subtyping_cor2:
  forall (Pi R T U: TypeS) (k: KindS),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi) ->
    let D := (Pi :: nil) in
    (TypeEnv_WK D) ->
    (Type_WK D T Type_) ->
    (Type_WK D U Type_) ->
    (Subtyping D T U) ->
    (Subtyping nil
               (typeSubst_in_type T 0 R)
               (typeSubst_in_type U 0 R))
.

Proof.
intros.
cut (
    (Subtyping (nil ++ (typeSubst_in_typeEnv nil 0 R))
               (typeSubst_in_type T (length (nil (A := TypeS))) R)
               (typeSubst_in_type U (length (nil (A := TypeS))) R))
).
simpl. tauto.
eapply Type_substitution_preserves_subtyping; eauto.
Qed.

(**********************************************************************)

Lemma typeSubst_of_some_inversion:
  forall (D: TypeEnv) (t N R: TypeS) (k: nat),
    (Type_WK D t TypeOption) ->
    (typeSubst_in_type t k R) = (Type_some N) ->
    exists N0: TypeS,
      t = (Type_some N0) /\
      N = (typeSubst_in_type N0 k R)
.

Proof.
intros. induction t; inversion H.
inversion H0.
simpl in H0. injection H0. intros. subst N. exists t; tauto.
Qed.

Lemma typeSubst_of_cons_inversion:
  forall (D: TypeEnv) (t u us R: TypeS) (k: nat),
    (Type_WK D t TypeParamList) ->
    (typeSubst_in_type t k R) = (Type_cons u us) ->
    exists u0: TypeS, exists us0: TypeS,
      t  = (Type_cons u0 us0) /\
      u  = (typeSubst_in_type u0 k R) /\
      us = (typeSubst_in_type us0 k R)
.

Proof.
intros. induction t; inversion H.
inversion H0.
simpl in H0. injection H0. intros. subst us. subst u.
exists t1; exists t2; tauto.
Qed.

Lemma TypeS_at_of_typeSubst_inversion:
  forall (D: TypeEnv) (t u R: TypeS) (i k: nat),
    (Type_WK D t TypeParamList) ->
    (TypeS_at (typeSubst_in_type t k R) i u) ->
    exists u0: TypeS,
      (TypeS_at t i u0) /\
      u = (typeSubst_in_type u0 k R)
.

Proof.
cut (
  forall (D: TypeEnv) (t' u R: TypeS) (i: nat),
    (TypeS_at t' i u) ->
    forall (t: TypeS) (k: nat),
      (Type_WK D t TypeParamList) ->
      t' = (typeSubst_in_type t k R) ->
      exists u0: TypeS,
        (TypeS_at t i u0) /\
        u = (typeSubst_in_type u0 k R)
).
intros. eapply H; eauto.
intros D t' u R i H. induction H.
(* *)
intros.
assert ((typeSubst_in_type t0 k R) = (Type_cons t ts)); auto. clear H0.
destruct (typeSubst_of_cons_inversion _ _ _ _ _ _ H H1).
destruct H0. destruct H0. destruct H2. subst t. subst ts. subst t0.
exists x. split. constructor. trivial.
(* *)
intros.
assert ((typeSubst_in_type t0 k R) = (Type_cons t ts)); auto. clear H1.
destruct (typeSubst_of_cons_inversion _ _ _ _ _ _ H0 H2).
destruct H1. destruct H1. destruct H3. subst t. subst ts. subst t0.
inversion_clear H0.
destruct (IHTypeS_at x0 k H3).  trivial. destruct H0.
exists x1. split. constructor; trivial. trivial.
Qed.

(************************************************************)

Lemma  Type_substitution_preserves_param_satisfaction:
  forall (D1 D2: TypeEnv) (Pi R K P: TypeS) (k k': KindS),
    let D := (D1 ++ (Pi :: D2)) in
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK D) ->
    (ParamSat D K P) ->
    (Type_WK D K (TypeConstr k')) ->
    (Type_WK D P TypeParam) ->
    (kind_of P k') ->
    (ParamSat (D1 ++ (typeSubst_in_typeEnv D2 0 R))
              (typeSubst_in_type K (length D2) R)
              (typeSubst_in_type P (length D2) R))
.

Proof.
intros.
destruct H2.
simpl. econstructor.
apply Substitution_preserves_removeBounds. eauto.
intros.
(* *)
inversion_clear H4.
destruct (typeSubst_of_some_inversion _ _ _ _ _ H9 H7).
destruct H4. rename x into N0. subst N.
(* *)
replace (snoc (D1 ++ typeSubst_in_typeEnv D2 0 R) Pi') with
  (D1 ++ (typeSubst_in_typeEnv (snoc D2 Pi') 0 R)).
Focus 2. rewrite snoc_of_append. rewrite typeSubst_of_snoc.
rewrite removeBounds_invariant_by_substitution with (t := Pi0). trivial. trivial.
(* *)
replace (Type_apply (lift_in_type (typeSubst_in_type K (length D2) R) 1)
                    (Type_var 0)) with
        (typeSubst_in_type (Type_apply (lift_in_type K 1) (Type_var 0))
                           (length (snoc D2 Pi'))
                           R).
Focus 2. simpl. rewrite length_of_snoc.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
rewrite Permuting_type_lifting_and_substitution_cor1. trivial.
omega.
(* *)
replace (typeSubst_in_type N0 (S (length D2)) R) with
        (typeSubst_in_type N0 (length (snoc D2 Pi')) R).
Focus 2. rewrite length_of_snoc. trivial.
(* *)
eapply Type_substitution_preserves_subtyping. eauto. eauto.
(* *)
rewrite <- snoc_of_cons. rewrite <- snoc_of_append.
constructor. trivial.
eapply removeBounds_preserves_well_kindedness. eauto. trivial.
(* *)
rewrite <- snoc_of_cons. rewrite <- snoc_of_append. auto.
(* *)
inversion H5. subst k'.
econstructor.
rewrite <- snoc_of_cons. rewrite <- snoc_of_append.
apply Weakening_of_well_kindedness_by_the_right_snoc. eauto.
econstructor. econstructor.
rewrite <- snoc_of_cons. rewrite <- snoc_of_append.
constructor.
unfold lift_in_type. apply Lifting_preserves_kind_of.
eapply removeBounds_preserves_kind_of. apply H13. trivial.
(* *)
subst N_opt. inversion_clear H9.
rewrite <- snoc_of_cons. rewrite <- snoc_of_append. unfold snoc.
assert (exists k0: KindS, (kind_of Pi0 k0)).
eapply wellKinded_implies_kind_of_exists; eauto. destruct H9. rename x into k0.
eapply Well_kindedness_does_not_depend_on_bounds. apply H9.
eapply removeBounds_preserves_kind_of. apply H9. trivial.
trivial.
Qed.

Theorem Type_substitution_preserves_satisfaction:
  forall (D1 D2: TypeEnv) (Pi R R' Pi': TypeS) (k: KindS),
    let D := (D1 ++ (Pi :: D2)) in
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK D) ->
    (SectionSat D R' Pi') ->
    (Type_WK D Pi' TypeSection) ->
    (SectionSat (D1 ++ (typeSubst_in_typeEnv D2 0 R))
               (typeSubst_in_type R' (length D2) R)
               (typeSubst_in_type Pi' (length D2) R))
.

Proof.
intros.
inversion H2. subst Pi'.
inversion H3. clear D0 Ps0 H8 H7.
unfold Forall_ParamSat in H6.
simpl. econstructor.
(* *)
apply Substitution_preserves_kind_of. eauto.
(* *)
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
(* *)
unfold Forall_ParamSat. intros.
(* *)
rewrite <- Permuting_type_substitutions_cor1 in H7.
assert (Type_WK D (typeSubst_in_type Ps 0 R') TypeParamList).
eapply Substitution_preserves_well_kindedness_snoc. eauto.
apply Kind_of_section. eauto. trivial.
destruct (TypeS_at_of_typeSubst_inversion _ _ _ _ _ _ H8 H7).
destruct H10. rename x into P0. subst P.
(* *)
change (Type_proj (typeSubst_in_type R' (length D2) R) i) with
       (typeSubst_in_type (Type_proj R' i) (length D2) R).
(* *)
assert (exists k : KindS, (KindS_at ks i k) /\ (kind_of P0 k)).
eapply kind_of_params_inversion_variant1.
apply Substitution_preserves_kind_of. apply H4. eauto.
destruct H11. destruct H11. rename x into k0.
(* *)
eapply Type_substitution_preserves_param_satisfaction. eauto. eauto.
trivial. auto.
(* *)
econstructor; eauto.
(* *)
eapply TypeParamList_WK_inversion. apply H8. eauto.
(* *)
trivial.
Qed.

Theorem Type_substitution_preserves_satisfaction_snoc:
  forall (D1: TypeEnv) (Pi R R' Pi': TypeS) (k: KindS),
    let D := (snoc D1 Pi) in
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK D) ->
    (SectionSat D R' Pi') ->
    (Type_WK D Pi' TypeSection) ->
    (SectionSat D1 (typeSubst_in_type R' 0 R) (typeSubst_in_type Pi' 0 R))
.

Proof.
intros.
replace
 ( SectionSat D1 (typeSubst_in_type R' 0 R) (typeSubst_in_type Pi' 0 R))
with
    (SectionSat (D1 ++ (typeSubst_in_typeEnv nil 0 R))
               (typeSubst_in_type R' (length (nil (A := TypeS))) R)
               (typeSubst_in_type Pi' (length (nil (A := TypeS))) R)).
eapply Type_substitution_preserves_satisfaction; eauto.
simpl. rewrite <- app_nil_end. trivial.
Qed.

End SetProgram.
