(* Author: Vincent Cremet *)
(* Date: December 2006 *) 
(* File: Substitutions_proofs.v *)
(* Description: Substitutions in FGJ-omega (Proofs) *)

(** * Definition of Substitutions (Proofs) *)

Require Import List.
Require Import Arith.

Require Import Util.
Require Import MatchCompNat.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Binders_proofs.v *)
Require Binders_proofs.
Module My_Binders_proofs := Binders_proofs.SetProgram(My_Program).
Import My_Binders_proofs.

Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(*****************************************************************)

(** Permuting type lifting and substitution (auxiliary) *)
Lemma Permuting_type_lifting_and_substitution_aux:
  forall (R: TypeS) (k1 n: nat) (t: TypeS) (k2: nat),
    (lift_at_depth_in_type (typeSubst_in_type t (k1 + k2) R) n k2) =
    (typeSubst_in_type (lift_at_depth_in_type t n k2) (k1 + k2 + n) R)
.

Proof.
intros R k1 n t. induction t.

(* proj *)
intro. simpl. rewrite IHt. trivial.

(* class *)
intro. simpl. trivial.

(* fun *)
intro. simpl. rewrite IHt1.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (k1 + k2 + n)) with (k1 + (S k2) + n). rewrite IHt2. trivial.
omega. auto with arith.

(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.

(* var *)
rename t into X.
intro. simpl.
induction (lt_eq_lt_dec X (k1 + k2)). induction a.
(* X < k1 + k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
simpl.
induction (match_comp_nat2_aux_0 X k2).
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial. omega. trivial.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_lt. trivial. omega. trivial. trivial.
(* X = k1 + k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_eq.
unfold lift_in_type.
pattern k2 at 2.
replace (k2) with (0 + k2).
rewrite Successive_type_liftings_aux. trivial. trivial.
omega. omega. trivial.
(* X > k1 + k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
simpl.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_gt.
replace (pred X + n) with (pred (X + n)). trivial.
omega. omega. omega. omega. trivial.

(* tuple *)
intro. simpl. rewrite IHt. trivial.

(* param *)
intro. simpl. rewrite IHt1.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (k1 + k2 + n)) with (k1 + (S k2) + n). rewrite IHt2. trivial.
omega. auto with arith.

(* section *)
intro. simpl.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (k1 + k2 + n)) with (k1 + (S k2) + n). rewrite IHt. trivial.
omega. auto with arith.

(* none *)
intro. simpl. trivial.

(* some *)
intro. simpl. rewrite IHt. trivial.

(* nil *)
intro. simpl. trivial.

(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Permuting_type_lifting_and_substitution:
  forall (t: TypeS) (X: TypeVar) (R: TypeS) (n: nat),
    (lift_in_type (typeSubst_in_type t X R) n) =
    (typeSubst_in_type (lift_in_type t n) (X + n) R)
.

Proof.
intros.
unfold lift_in_type.
replace (X) with (X + 0).
apply Permuting_type_lifting_and_substitution_aux.
auto with arith.
Qed.

Lemma Permuting_type_lifting_and_substitution_cor1:
  forall (t : TypeS) (X : TypeVar) (R : TypeS),
    (lift_in_type (typeSubst_in_type t X R) 1) =
    (typeSubst_in_type (lift_in_type t 1) (S X) R)
.

Proof.
intros.
replace (S X) with (X + 1).
apply Permuting_type_lifting_and_substitution. omega.
Qed.

(************************************************************)

Lemma Permuting_type_liftings_aux:
  forall (k1 n1 n2: nat) (t: TypeS) (k2: nat),
    (lift_at_depth_in_type (lift_at_depth_in_type t n2 k2) n1 (k1 + n2 + k2)) =
    (lift_at_depth_in_type (lift_at_depth_in_type t n1 (k1 + k2)) n2 k2)
.

Proof.
intros k1 n1 n2 t. induction t.
(* proj *)
intro. simpl. rewrite IHt. trivial.
(* class *)
intro. simpl. trivial.
(* fun *)
intro. simpl. rewrite IHt1.
replace (S (k1 + n2 + k2)) with (k1 + n2 + (S k2)).
replace (S (k1 + k2)) with (k1 + (S k2)). rewrite IHt2. trivial.
auto with arith. auto with arith.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* var *)
rename t into X.
intro. simpl.
unfold lift_at_depth_in_typeVar.
induction (match_comp_nat2_aux_0 X k2).
(* X < k2 *)
rewrite (match_comp_nat2_lt X k2).
rewrite match_comp_nat2_lt.
rewrite (match_comp_nat2_lt X (k1 + k2)).
rewrite match_comp_nat2_lt. trivial.
trivial. omega. omega. trivial.
(* X >= k2 *)
rewrite (match_comp_nat2_ge X k2).
induction (match_comp_nat2_aux_0 X (k1 + k2)).
(* X < k1 + k2 *)
rewrite match_comp_nat2_lt.
rewrite (match_comp_nat2_lt X (k1 + k2)).
rewrite match_comp_nat2_ge. trivial.
trivial. trivial. omega.
(* X >= k1 + k2 *)
rewrite match_comp_nat2_ge. 
rewrite (match_comp_nat2_ge X (k1 + k2)).
rewrite match_comp_nat2_ge. replace (X + n2 + n1) with (X + n1 + n2). trivial.
omega. omega. trivial. omega. trivial.
(* tuple *)
intro. simpl. rewrite IHt. trivial.
(* param *)
intro. simpl. rewrite IHt1.
replace (S (k1 + n2 + k2)) with (k1 + n2 + (S k2)).
replace (S (k1 + k2)) with (k1 + (S k2)). rewrite IHt2. trivial.
auto with arith. auto with arith.
(* section *)
intro. simpl.
replace (S (k1 + n2 + k2)) with (k1 + n2 + (S k2)).
replace (S (k1 + k2)) with (k1 + (S k2)). rewrite IHt. trivial.
auto with arith. auto with arith.
(* none *)
intro. simpl. trivial.
(* some *)
intro. simpl. rewrite IHt. trivial.
(* nil *)
intro. simpl. trivial.
(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Permuting_type_liftings_cor1:
  forall (k1 n1 n2: nat) (t: TypeS),
    (n2 <= k1) ->
    (lift_at_depth_in_type (lift_in_type t n2) n1 k1) =
    (lift_in_type (lift_at_depth_in_type t n1 (k1 - n2)) n2)
.

Proof.
intros.
cut (
    (lift_at_depth_in_type (lift_at_depth_in_type t n2 0) n1 ((k1 - n2) + n2 + 0)) =
    (lift_at_depth_in_type (lift_at_depth_in_type t n1 ((k1 - n2) + 0)) n2 0)
).
unfold lift_in_type.
replace (k1 - n2 + n2 + 0) with k1.
replace (k1 - n2 + 0) with (k1 - n2). trivial. omega. omega.
apply Permuting_type_liftings_aux.
Qed.

(** Permuting type lifting and substitution 2 (auxiliary) *)
Lemma Permuting_type_lifting_and_substitution_2_aux:
  forall (R: TypeS) (k1 n: nat) (t: TypeS) (k2: nat),
    (lift_at_depth_in_type (typeSubst_in_type t k2 R) n (k1 + k2)) =
    (typeSubst_in_type (lift_at_depth_in_type t n (S (k1 + k2))) k2 
      (lift_at_depth_in_type R n k1))
.

Proof.
intros R k1 n t. induction t.
(* proj *)
intro. simpl. rewrite IHt. trivial.
(* class *)
intro. simpl. trivial.
(* fun *)
intro. simpl. rewrite IHt1.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (S (k1 + k2))) with (S (k1 + (S k2))).
rewrite IHt2. trivial. auto with arith. auto with arith.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* var *)
rename t into X.
intro. simpl.
induction (lt_eq_lt_dec X k2). induction a.
(* X < k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
simpl. unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt. rewrite match_comp_nat3_lt. trivial.
trivial. omega. omega. trivial.
(* X = k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_eq. unfold lift_in_type.
replace (k1 + k2) with (k1 + k2 + 0).
pattern k1 at 2. replace (k1) with (k1 + 0).
rewrite Permuting_type_liftings_aux. trivial.
auto with arith. auto with arith.
trivial. omega. trivial.
(* X > k2 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
induction (match_comp_nat2_aux_0 X (S (k1 + k2))).
(* X < k1 + k2 + 1 *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_gt.
simpl. unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt. trivial.
omega. trivial. trivial.
(* X >= k1 + k2 + 1 *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
simpl. unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_gt. replace (pred X + n) with (pred (X + n)). trivial.
omega. omega. omega. trivial. trivial.
(* tuple *)
intro. simpl. rewrite IHt. trivial.
(* param *)
intro. simpl. rewrite IHt1.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (S (k1 + k2))) with (S (k1 + (S k2))).
rewrite IHt2. trivial. auto with arith. auto with arith.
(* section *)
intro. simpl.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (S (k1 + k2))) with (S (k1 + (S k2))).
rewrite IHt. trivial. auto with arith. auto with arith.
(* none *)
intro. simpl. trivial.
(* some *)
intro. simpl. rewrite IHt. trivial.
(* nil *)
intro. simpl. trivial.
(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Permuting_type_lifting_and_substitution_2_cor1:
  forall (R : TypeS) (k n : nat) (t : TypeS),
    (lift_at_depth_in_type (typeSubst_in_type t 0 R) n k) =
    (typeSubst_in_type (lift_at_depth_in_type t n (S k)) 0
                       (lift_at_depth_in_type R n k))
.

Proof.
intros.
cut (
   (lift_at_depth_in_type (typeSubst_in_type t 0 R) n (k + 0)) =
    (typeSubst_in_type (lift_at_depth_in_type t n (S (k + 0))) 0
                       (lift_at_depth_in_type R n k))
).
rewrite plus_comm. tauto.
apply Permuting_type_lifting_and_substitution_2_aux.
Qed.

(************************************************************************)

Lemma Type_lifting_stronger_than_substitution_aux:
  forall (n d: nat) (R t: TypeS) (k: nat),
    (typeSubst_in_type (lift_at_depth_in_type t (S (n + d)) k) (n + k) R) =
    (lift_at_depth_in_type t (n + d) k)
.

Proof.
intros n d R t. induction t.

(* prof *)
intro. simpl. rewrite IHt. trivial.
(* class *)
intro. simpl. trivial.
(* fun *)
intro. simpl. rewrite IHt1.
replace (S (n + k)) with (n + (S k)). rewrite IHt2. trivial.
auto with arith.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* var *)
rename t into X.
intro. simpl.
induction (match_comp_nat2_aux_0 X k).
(* X < k *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt. trivial.
omega. trivial. trivial.
(* X >= k *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
replace (pred (X + S (n + d))) with (X + (n + d)). trivial.
omega. omega. trivial. trivial.
(* tuple *)
intro. simpl. rewrite IHt. trivial.
(* param *)
intro. simpl. rewrite IHt1.
replace (S (n + k)) with (n + (S k)). rewrite IHt2. trivial.
auto with arith.
(* section *)
intro. simpl.
replace (S (n + k)) with (n + (S k)). rewrite IHt. trivial.
auto with arith.
(* none *)
intro. simpl. trivial.
(* some *)
intro. simpl. rewrite IHt. trivial.
(* nil *)
intro. simpl. trivial.
(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Type_lifting_stronger_than_substitution:
  forall (n d: nat) (R t: TypeS),
    (typeSubst_in_type (lift_in_type t (S (n + d))) n R) =
    (lift_in_type t (n + d))
.

Proof.
intros.
unfold lift_in_type.
pattern n at 2. replace (n) with (n + 0).
apply Type_lifting_stronger_than_substitution_aux.
auto with arith.
Qed.

(*************************************************************************)

(** Permuting type substitutions *)
Lemma Permuting_type_substitutions:
  forall (R R': TypeS) (k: nat) (t: TypeS) (n: nat),
    (typeSubst_in_type (typeSubst_in_type t n R) (n + k) R') =
    (typeSubst_in_type (typeSubst_in_type t (S (n + k)) R') n (typeSubst_in_type R k R'))
.

Proof.
intros R R' k t.
induction t.

(* proj *)
intro. simpl.
rewrite IHt. trivial.

(* class *)
intro. simpl. trivial.

(* fun *)
intro. simpl.
rewrite IHt1.
replace (S (n + k)) with ((S n) + k).
rewrite IHt2. trivial. trivial.

(* apply *)
intro. simpl.
rewrite IHt1. rewrite IHt2. trivial.

(* var *)
rename t into X.
intro. simpl.
induction (lt_eq_lt_dec X n). induction a.
(* X < n *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt. simpl.  unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
trivial.
trivial. auto with arith. auto with arith. trivial.
(* X = n *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_lt.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
rewrite Permuting_type_lifting_and_substitution.
replace (n + k) with (k + n). trivial. auto with arith.
trivial. omega. trivial.
(* X > n *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
simpl.
induction (lt_eq_lt_dec X (S (n + k))). induction a.
(* X < n + k + 1 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
trivial.
trivial. trivial. omega.
(* X = n + k + 1 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_eq.
rewrite Type_lifting_stronger_than_substitution. trivial.
trivial. omega.
(* X > n + k + 1 *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
rewrite match_comp_nat3_gt.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
trivial. omega. trivial. omega. trivial.

(* tuple *)
intro. simpl. rewrite IHt. trivial.

(* param *)
intro. simpl. rewrite IHt1.
replace (S (n + k)) with ((S n) + k).
rewrite IHt2. trivial. trivial.

(* section *)
intro. simpl.
replace (S (n + k)) with ((S n) + k).
rewrite IHt. trivial. trivial.

(* none *)
intro. simpl. trivial.

(* some *)
intro. simpl. rewrite IHt. trivial.

(* nil *)
intro. simpl. trivial.

(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Permuting_type_substitutions_cor1:
  forall (R R': TypeS) (k: nat) (t: TypeS),
    (typeSubst_in_type (typeSubst_in_type t 0 R) k R') =
    (typeSubst_in_type (typeSubst_in_type t (S k) R') 0 (typeSubst_in_type R k R'))
.

Proof.
intros.
cut(
    (typeSubst_in_type (typeSubst_in_type t 0 R) (0 + k) R') =
    (typeSubst_in_type (typeSubst_in_type t (S (0 + k)) R') 0 (typeSubst_in_type R k R'))
).
simpl. trivial. apply Permuting_type_substitutions.
Qed.

Lemma Permuting_type_substitutions_cor2:
  forall (R R': TypeS) (k: nat) (t: TypeS),
    (typeSubst_in_type (typeSubst_in_type t 1 R) (S k) R') =
    (typeSubst_in_type (typeSubst_in_type t (S (S k)) R') 1 (typeSubst_in_type R k R'))
.

Proof.
intros.
cut(
    (typeSubst_in_type (typeSubst_in_type t 1 R) (1 + k) R') =
    (typeSubst_in_type (typeSubst_in_type t (S (1 + k)) R') 1 (typeSubst_in_type R k R'))
).
simpl. trivial. apply Permuting_type_substitutions.
Qed.

End SetProgram.
