(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: TypeSafety.v *)
(* Description: Type safety *)

(** * Type safety *)

Require Import List.
(* Require Import Arith. *)
Require Import Relations.

(* Require Import MatchCompNat. *)
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of SubjectReduction_proofs.v *)
Require SubjectReduction_proofs.
Module My_SubjectReduction_proofs := SubjectReduction_proofs.SetProgram(My_Program).
Import My_SubjectReduction_proofs.

Import My_SubjectReduction_preliminaries.
Import My_TermSubstitutionInTyping.
Import My_TypeSubstitutionInTyping.
Import My_BoundCompliantSubstitutions_proofs.
Import My_Progress_proofs.
Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

(** Reflexive transitive closure of the reduction relation **)
Definition RedTermStar :=
  (clos_refl_trans _ RedTerm)
.

(********************************************************************)

(** Multi-step subject reduction **)
Lemma Subject_reduction_star:
  forall (t u: Term) (info: TermInfo),
    (Term_WT nil nil t info) ->
    (RedTermStar t u) ->
    exists info': TermInfo,
      (Term_WT nil nil u info') /\
      (SubInfo nil info' info)
.

Proof.
intros. generalize info H. clear info H. induction H0.
(* step *)
intros.
destruct (Subject_reduction _ _ _ H0 H). destruct H1. exists x0. tauto.
(* refl *)
intros.
exists info. split. trivial. apply SubInfo_refl.
(* *)
intros.
destruct (IHclos_refl_trans1 _ H). destruct H0. rename x0 into info'.
destruct (IHclos_refl_trans2 _ H0). destruct H2. rename x0 into info''.
exists info''. split.
trivial.
eapply SubInfo_trans; eauto.
eapply Typing_returns_well_kinded_types; eauto.
apply List_forall_nil.
Qed.

Lemma Subject_reduction_star_cor1:
  forall (t t': Term) (T: TypeS),
    (Term_WT nil nil t (TermI T)) ->
    (RedTermStar t t') ->
    exists T': TypeS,
      (Term_WT nil nil t' (TermI T'))
.

Proof.
intros.
destruct (Subject_reduction_star _ _ _ H H0). destruct H1. inversion H2.
rewrite <- H4 in H1.
exists T0. tauto.
Qed.

(********************************************************************)

(** Type safety **)
Lemma Type_safety:
  forall (t: Term),
    (RedTermStar mainTerm t) ->
    (evaluated_term t) \/ (exists u: Term, (RedTerm t u))
.

Proof.
intros.
destruct (main_term_is_well_typed). rename x into T.
destruct (evaluated_term_dec t).
left. trivial.
right.
destruct (Subject_reduction_star_cor1 _ _ _ H0 H).
eapply Progress; eauto.
Qed.

End SetProgram.
